
# coding: utf-8

# # Independent Component Analysis
# 
# ### Laurenz Wiskott (2016-09-03)
# 
# In this notebook two independent source signals are mixed and then separated based on fourth order cumulants.

# ## Imports

# In[1]:


import numpy as np
import matplotlib.pyplot as plt
import math as m


# ## Generate mixed data and plot it

# In[2]:


numDat = 1000                                 # Number of data points. Play with this.
dimDat = 2                                    # Dimensionality of data point. Do not change!


# The sources are sampled from a uniform and a Laplace distribution with variance one, then mixed with a random mixing matrix and offset by a random mean to generate the mixed data.  Coefficients of mean vector and mixing matrix are drawn from a normal distribution with variance one.

# In[3]:


source1 = np.random.uniform(-np.sqrt(3),np.sqrt(3),(numDat))  # Sample 1st source from uniform distribution with variance one. 
source2 = np.random.laplace(0,1/np.sqrt(2),(numDat))          # Sample 2nd source from Laplace distribution with variance one.
sources = np.vstack([source1,source2])                        # Combine first and second source.

meanInit = np.random.normal(0,1,dimDat)                       # Initialize mean from normal distribution.
mixingMatrix = np.random.normal(0,1,(dimDat,dimDat))          # Create random mixing matrix.
data = np.dot(mixingMatrix,sources) + meanInit.reshape((2,1)) # Mix initial sources and add mean.

print "\nInitial sources:\n", sources
print "\nInitial mean:\n", meanInit
print "\nMixing matrix:\n", mixingMatrix
print "\nMixed data:\n", data


# In[4]:


fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111)
plotRange = 4*np.max(np.absolute(mixingMatrix))
ax.set(xlim=[meanInit[0]-plotRange,meanInit[0]+plotRange], ylim=[meanInit[1]-plotRange,meanInit[1]+plotRange], title='Mixed Data', xlabel='First Component', ylabel='Second Component')
ax.scatter(data[0],data[1],s=3) # Plot data points.
plt.show()


# ## Whiten the data and plot it
# 
# The first step in separating the sources is to whiten the data, so that projecting it onto two orthogonal axes yields uncorrelated signals.
# 
# Hints: `np.average`, `np.dot`, `np.transpose`, `np.linalg.eig`, and `np.sqrt might be useful here.  Use `print np.<name>?` to learn more about `<name>`.

# In[5]:


# np.average?


# In[6]:


## EXERCISE 1

# Add missing code here.
# ...
# dataMean = ...        # Data mean as a 2 dimensional vector.
# whiteningMatrix = ... # Whitening matrix as a 2 x 2 dimensional array.
# whitenedData = ...    # Whitened data as a 2 x numDat dimensional array.

# SOLUTION

dataMean = np.average(data,axis=1)                               # Calculate data mean.
zeroMeanData = data - dataMean.reshape((2,1))                    # Calculate zero mean data.
print "\nNumerical mean of zero mean data (should be zero):\n", np.average(zeroMeanData,axis=1) 
                                                                 # Verify that data has zero mean.
secMomMat = np.dot(zeroMeanData,zeroMeanData.transpose())/numDat # Calculate 2nd moment matrix.
print "\nSecond moment matrix:\n", secMomMat
pcVar, pcVec = np.linalg.eig(secMomMat)                          # Calculate eigenvalues and eigenvectors.
print "\nPrincipal components:\n", pcVec
print "\nVariances:\n", pcVar
whiteningMatrix = np.dot(np.dot(pcVec,np.diagflat(1/np.sqrt(pcVar))),pcVec.transpose()) 
                                                                 # Calculate whitening matrix from eigenvectors and eigenvalues.
whitenedData = np.dot(whiteningMatrix, zeroMeanData)             # Calculate whitened data by multiplying zero mean data with whitening matrix.

# EXERCISE END

print "\nData mean:\n", dataMean
print "\nWhitening matrix:\n", whiteningMatrix
print "\nWhitened data:\n", whitenedData


# In[7]:


fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111)
plotRange = 4
ax.set(xlim=[-plotRange,+plotRange], ylim=[-plotRange,+plotRange], title='Whitened Data', xlabel='First Component', ylabel='Second Component')
ax.scatter(whitenedData[0],whitenedData[1],s=3) # Plot data points.
plt.show()


# ## Define function for calculating fourth order autocumulants for zero mean data
# 
# The contrast function is based on fourth order auto cumulants, which are calculated with the following function.

# In[8]:


def fourthOrderAutoCumulants(zeroMeanData):
    """Calculate 4th order autocumulants for zero mean data.
    Input: data array (dimDat x numDat).
    Outpu: forth orther autocumulant array (dimDat)."""
    
    ## EXERCISE 2
    
    # Add missing code here.
    # ...
    # cumulants4thOrder = ...
    
    # SOLUTION
    
    moment2ndOrder = np.average(zeroMeanData**2,axis=1)
    moment4thOrder = np.average(zeroMeanData**4,axis=1)
    cumulants4thOrder = moment4thOrder - 3*moment2ndOrder**2
    
    # EXERCISE END
    
    return cumulants4thOrder


# In[9]:


print fourthOrderAutoCumulants(whitenedData)


# ## Define contrast function for whitened data
# 
# The contrast function is a scalar function that quantifies on the basis of fourth order cumulants how non-Gaussian the components of a signal are. 

# In[10]:


def contrastFunction(whitenedData):
    """Calculate ICA contrast function for whitened data (shall be maximized).
    Input: whitened data array (dimDat x numDat).
    Output: contrast function value (float)."""
    
    ## EXERCISE 3
    
    # Add missing code here.
    # ...
    # contrast = ...
    
    # SOLUTION
    
    cumulants = fourthOrderAutoCumulants(whitenedData)
    contrast = np.sum(cumulants**2)
    
    # EXERCISE END
    
    return contrast


# In[11]:


print contrastFunction(whitenedData)


# ## Find optimal rotation angle with simple method
# 
# This is a simple procedure to rotate the data such that the contrast function is maximized, i.e. such that the components are maximally non-Gaussian, which implies that the components are minimally statistically independent (as far as fourth order is concerned).

# In[12]:


# Start with rotation angle 0, i.e. with rotation matrix [[1,0],[0,1]] and contrast of the whitened data.
bestAngle = 0
bestRotationMatrix = np.array([[1,0],[0,1]])
bestRotatedContrast = contrastFunction(whitenedData)

angle = 0
for iteration in np.arange(0,30,0.01):

    # Try a new rotation angle near the old one.
    angle = angle + np.random.normal(0,m.exp(-iteration))    # Vary the angle a bit, where 'a bit' gets exponentially smaller.
    rotationMatrix = np.array([[m.cos(angle),-m.sin(angle)],[m.sin(angle),m.cos(angle)]]) # Calculate new rotation matrix.
    rotatedData = np.dot(rotationMatrix,whitenedData)        # Calculate rotated whitened data.
    rotatedContrast = contrastFunction(rotatedData)          # Calculate contrast ov newly rotated whitened data.
    
    # If contrast has improved, keep the new rotation angle, matrix, and contrast, otherwise continue with the old ones.
    if rotatedContrast > bestRotatedContrast:
        bestAngle = angle
        print m.cos(angle), m.sin(angle)                     # Print the cos and sin of the improved angle to demonstrate progress (the angle is not so useful, since it has a 2pi periodicity, and the problem has a 90° symmetry)
        bestRotationMatrix = rotationMatrix
        bestRotatedContrast = rotatedContrast
        
# Display final result.
print "Best angle: ", bestAngle
print "Best rotation matrix:\n", bestRotationMatrix
print "Best rotated contrast: ", bestRotatedContrast
    


# ## Calculate unmixing matrix and plot separated sources
# 
# Once one has the optimal rotation matrix for the whitened data, one can easily calculate an unmixing matrix that can be applied to the mixed data to separate the sources.

# In[13]:


## EXERCISE 4
    
# Add missing code here.
# unmixingMatrix = ...
    
# SOLUTION

unmixingMatrix = np.dot(bestRotationMatrix,whiteningMatrix)

# EXERCISE END

separatedSources = np.dot(unmixingMatrix,zeroMeanData)


# In[14]:


fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111)
plotRange = 4
ax.set(xlim=[-plotRange,+plotRange], ylim=[-plotRange,+plotRange], title='Separated Sources', xlabel='First Source', ylabel='Second Source')
ax.scatter(separatedSources[0],separatedSources[1],s=3) # Plot data points.
plt.show()


# ## Plot mixed data with mixing and unmixing vectors

# The following plot shows the mixed data points in black, the true mean offset vector in yellow, the data mean vector in blue on top, the mixing vectors in green, and the unmixing vectors in red on top.  The mixing vectors are the column vectors of the mixing matrix, and the unmixing vectors are the row vectors of the unmixing matrix.  Since the polarity of the signals is arbitrary, mixing and unmixing vectors are plotted $\pm$.

# In[15]:


fig = plt.figure(figsize=(12,12))
ax = fig.add_subplot(111)
plotRange = 4*np.max(np.absolute(mixingMatrix))
ax.set(xlim=[dataMean[0]-plotRange,dataMean[0]+plotRange], ylim=[dataMean[1]-plotRange,dataMean[1]+plotRange], title='Mixed Data', xlabel='First Component', ylabel='Second Component')
ax.scatter(data[0],data[1],s=3)          # Plot data points in black.
ax.plot([0,meanInit[0]],[0,meanInit[1]],
        'yo-',markersize=15,linewidth=5) # Plot ground truth mean offset of the mixed data in yellow.
for source in range(0,2):
    ax.plot(meanInit[0]+mixingMatrix[0,source]*np.array((-1,+1)),meanInit[1]+mixingMatrix[1,source]*np.array((-1,+1)),
        'go-',markersize=15,linewidth=5) # Plot +- mixing vectors in green.
ax.plot([0,dataMean[0]],[0,dataMean[1]],
        'bo-',markersize=15,linewidth=5) # Plot data mean offset of the mixed data in blue.
for source in range(0,2):
    ax.plot(dataMean[0]+unmixingMatrix[source,0]*np.array((-1,+1)),dataMean[1]+unmixingMatrix[source,1]*np.array((-1,+1)),
        'ro-',markersize=15,linewidth=5) # Plot +- mixing vectors in red.
plt.show()


# ## Exercise 5
# 
# What geometrical relationship should there be between the mixing vectors and the unmixing vectors in terms of position, angle and length?

# ### Solution
# 
# The position of all vectors should be at the true mean of the data distribution.
# 
# Generally the unmixing vectors should be orthogonal to the mixing vectors of those sources that they do not extract.  Thus, in higher dimensions the unmixing vectors should be orthogonal to all but one mixing vector.
# 
# Since the true sources as well as the extracted sources are normalized to variance one, any scaling done by the mixing vectors should be undone by a scaling of the corresponding unmixing vectors.  Thus for long mixing vectors the unmixing vectors should be short and the other was around.  More precisely, the inner product between a mixing vector and an unmixing vector should be 0 or 1 depending on whether they have a different index or same index, respectively.  Thus for a long mixing vector the one unmixing that is not orthogonal to it, should be short, though the angle between the two vectors also enters the equation.
# 

# ## Exercise 6
# 
# How can you formally check whether your unmixing matrix is correct?

# ### Solution
# 
# Ideally the unmixing matrix $\textbf{U}$ is the inverse of the mixing matrix $\textbf{M}$, and thus $\textbf{U}\textbf{M}=\textbf{1}$.  However ICA can only be expected to work up to a permutation and different scaling of the sources, thus $\textbf{U}\textbf{M}$ should look like a permutation matrix with arbitrary (but significantly different from zero) numbers instead of the ones, i.e. each row and each column should contain exactly one entry significantly different from 0.  However, since in our case the true sources as well as the estimated sources are normalized to variance one, the non-zero elements should be close to $\pm1$.

# In[16]:


print "Product of unmixing and mixing matrix (coefficients should be close to 0 and -1/+1):\n", np.dot(unmixingMatrix,mixingMatrix)


# ## Exercise 7
# 
# Vary the number of data points and observe how much the solution depends on it.

# ### Solution
# 
# I find it amazing how robust the solution is to the number of data points.  In some sense the mean seams more sensitive than the unmixing vectors.

# #####  This work is licensed under a <a rel="license" href="http://creativecommons.org/licenses/by/4.0/">Creative Commons Attribution 4.0 International License</a>.
